import operator
import pickle
from collections import OrderedDict
from collections.abc import Iterable, Iterator
from itertools import chain, islice
from os import PathLike
from pathlib import Path
from types import TracebackType
from typing import Any, Final, Generic, Optional, SupportsIndex, Type
from typing import TypeVar, Union, overload

from .index import NestedIndex
from .mutable_sequence_islice import MutableSequenceIslice
from .viewable_mutable_sequence import ViewableMutableSequence

ET = TypeVar("ET", bound=BaseException)
T = TypeVar("T")

Self = TypeVar("Self", bound="BigList")

CHUNKSIZE = 1024
CHUNKSIZE_EXTENDED = 1536

assert CHUNKSIZE * 3 == CHUNKSIZE_EXTENDED * 2

def ensure_file(path: Path, default: T) -> T:
    if path.exists():
        with open(path, mode="rb") as file:
            return pickle.load(file)
    else:
        with open(path, mode="wb") as file:
            pickle.dump(default, file)
        return default


class BigList(ViewableMutableSequence[T], Generic[T]):
    _cache: Final[OrderedDict[int, list[T]]]
    _filenames: Final[list[int]]
    _index: Final[NestedIndex]
    _path: Final[Path]

    __slots__ = {
        "_cache":
            "A cache containing the 4 most recently accessed segments.",
        "_filenames":
            "The file names for each segment.",
        "_index":
            "A data structure for fast repeated indexing.",
        "_path":
            "The folder containing all of the files.",
    }

    def __init__(self: Self, path: Union[PathLike, bytes, str], /) -> None:
        path = Path(path).resolve()
        path.mkdir(exist_ok=True)
        (path / "list").mkdir(exist_ok=True)
        ensure_file(path / "list" / "counter.txt", 0)
        self._cache = OrderedDict()
        self._filenames = ensure_file(path / "list" / "filenames.txt", [])
        self._index = NestedIndex(ensure_file(path / "list" / "lens.txt", []))
        self._path = path

    def __delitem__(self: Self, index: Union[int, slice], /) -> None:
        if isinstance(index, slice):
            len_ = len(self)
            range_ = range(len_)[index]
            size = len(range_)
            if range_.step < 0:
                range_ = range_[::-1]
            if size == len_:
                self.clear()
                return
            elif size == 0:
                return
            elif range_.step == 1 and range_.start == 0:
                i, j = self._index.find(range_.stop)
                for k in range(i):
                    (self._path / "list" / f"{self._filenames[k]}.txt").unlink()
                del self._filenames[:i]
                del self._index[:i]
                del self._cache_chunk(0)[:j]
                self._index.update(0, -j)
                self._balance(0)
            elif range_.step == 1 and range_.stop == len(self):
                i, j = self._index.find(range_.start)
                for k in range(i + 1, len(self._filenames)):
                    (self._path / "list" / f"{self._filenames[k]}.txt").unlink()
                del self._filenames[i + 1 :]
                del self._index[i + 1 :]
                del self._cache_chunk(-1)[j:]
                self._index[-1] = j
                self._balance(-1)
            elif range_.step == 1:
                start = self._index.find(range_.start)
                stop = self._index.find(range_.stop)
                del self._cache_chunk(stop[0])[:stop[1]]
                self._index.update(stop[0], -stop[1])
                for k in range(start[0] + 1, stop[0]):
                    (self._path / "list" / f"{self._filenames[k]}.txt").unlink()
                del self._filenames[start[0] + 1 : stop[0]]
                del self._index[start[0] + 1 : stop[0]]
                del self._cache_chunk(start[0])[start[1]:]
                self._index[start[0]] = start[1]
                self._balance(start[0])
            else:
                for i in reversed(range_):
                    del self[i]
        else:
            index = range(len(self))[index]
            i, j = self._index.find(index)
            del self._cache_chunk(i)[j]
            self._index.update(i, -1)
            self._balance(i)

    def __enter__(self: Self, /) -> Self:
        return self

    def __exit__(
        self: Self,
        exc_type: Optional[Type[ET]],
        exc_val: Optional[ET],
        exc_traceback: Optional[TracebackType],
        /,
    ) -> None:
        self.commit()

    @overload
    def __getitem__(self: Self, index: int, /) -> T: ...

    @overload
    def __getitem__(self: Self, index: slice, /) -> MutableSequenceIslice[T]: ...

    def __getitem__(self, index, /):
        if isinstance(index, slice):
            range_ = range(len(self))[index]
            return self.islice[index]
        else:
            index = range(len(self))[index]
            i, j = self._index.find(index)
            return self._cache_chunk(i)[j]

    def __getstate__(self: Self, /) -> Path:
        return self._path

    def __islice(self: Self, start: int, stop: int, step: int, /) -> Iterator[T]:
        f_start = self._index.find(start)
        f_stop = self._index.find(stop)
        if step > 0:
            n = f_start[1]
            for i in range(f_start[0], f_stop[0]):
                chunk = self._cache_chunk(i)
                while n < len(chunk):
                    yield chunk[n]
                    n += step
                n -= len(chunk)
            chunk = self._cache_chunk(f_stop[0])
            for i in range(n, f_stop[1] + 1, step):
                yield chunk[i]
        else:
            chunk = self._cache_chunk(f_start[0])
            n = f_start[1] - len(chunk)
            for i in reversed(range(f_stop[0], f_start[0])):
                while n > -len(chunk):
                    yield chunk[n]
                    n += step
                n += len(chunk)
                chunk = self._cache_chunk(i)
            for i in range(n, f_stop[1] - len(chunk) - 1, step):
                yield chunk[i]

    def __islice__(self: Self, start: Optional[int], stop: Optional[int], step: Optional[int], /) -> Iterator[T]:
        range_ = range(len(self))[start:stop:step]
        if len(range_) == len(self):
            return iter(self) if range_.step == 1 else reversed(self)
        elif len(range_) == 0:
            return (self[i] for i in range_)
        elif range_.step == 1 and range_.start == 0:
            return islice(self, range_.stop)
        elif range_.step == 1 and range_.stop == len(self):
            f_start = self._index.find(range_.start)
            if f_start[1] == 0:
                return chain.from_iterable(
                    self._cache_chunk(i)
                    for i in range(f_start[0], len(self._filenames))
                )
            else:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]:],
                    chain.from_iterable(
                        self._cache_chunk(i)
                        for i in range(f_start[0] + 1, len(self._filenames))
                    ),
                )
        elif range_.step == 1:
            f_start = self._index.find(range_.start)
            f_stop = self._index.find(range_.stop)
            if f_start[0] == f_stop[0]:
                return iter(self._cache_chunk(f_start[0])[f_start[1]:f_stop[1]])
            elif f_start[1] == 0 == f_stop[1]:
                return chain.from_iterable(
                    self._cache_chunk(i)
                    for i in range(f_start[0], f_stop[0])
                )
            elif f_start[1] == 0:
                return chain(
                    chain.from_iterable(
                        self._cache_chunk(i)
                        for i in range(f_start[0], f_stop[0])
                    ),
                    self._cache_chunk(f_stop[0])[:f_stop[1]],
                )
            elif f_stop == 0:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]:],
                    chain.from_iterable(
                        self._cache_chunk(i)
                        for i in range(f_start[0] + 1, f_stop[0])
                    ),
                )
            else:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]:],
                    chain.from_iterable(
                        self._cache_chunk(i)
                        for i in range(f_start[0] + 1, f_stop[0])
                    ),
                    self._cache_chunk(f_stop[0])[:f_stop[1]],
                )
        elif range_.step == -1 and range_.start + 1 == len(self._filenames):
            return islice(reversed(self), len(range_))
        elif range_.step == -1 and range_.stop + 1 == 0:
            f_start = self._index.find(range_.start)
            if f_start[1] + 1 == self._index[f_start[0]]:
                return chain.from_iterable(reversed(self._cache_chunk(i)) for i in range(f_start[0], -1, -1))
            else:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]::-1],
                    chain.from_iterable(
                        reversed(self._cache_chunk(i))
                        for i in reversed(range(f_start[0]))
                    ),
                )
        elif range_.step == -1:
            f_start = self._index.find(range_.start)
            f_stop = self._index.find(range_.stop)
            if f_start[0] == f_stop[0]:
                return iter(self._cache_chunk(f_start[0])[f_start[1]:f_stop[1]:-1])
            elif (
                f_start[1] + 1 == self._index[f_start[0]]
                and f_stop[1] + 1 == self._index[f_stop[0]]
            ):
                return chain.from_iterable(
                    reversed(self._cache_chunk(i))
                    for i in range(f_start[0], f_stop[0], -1)
                )
            elif f_start[1] + 1 == self._index[f_start[0]]:
                return chain(
                    chain.from_iterable(
                        reversed(self._cache_chunk(i))
                        for i in range(f_start[0], f_stop[0], -1)
                    ),
                    self._cache_chunk(f_stop[0])[:f_stop[1]:-1],
                )
            elif f_stop + 1 == self._index[f_stop[0]]:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]::-1],
                    chain.from_iterable(
                        reversed(self._cache_chunk(i))
                        for i in range(f_start[0] - 1, f_stop[0], -1)
                    ),
                )
            else:
                return chain(
                    self._cache_chunk(f_start[0])[f_start[1]::-1],
                    chain.from_iterable(
                        reversed(self._cache_chunk(i))
                        for i in range(f_start[0] - 1, f_stop[0], -1)
                    ),
                    self._cache_chunk(f_stop[0])[:f_stop[1]:-1],
                )
        elif abs(range_.step) < CHUNKSIZE * 2:
            return self.__islice(range_.start, range_.stop, range_.step)
        else:
            return super().__islice__(range_.start, range_.stop, range_.step)

    def __iter__(self: Self, /) -> Iterator[T]:
        return chain.from_iterable(
            self._cache_chunk(i)
            for i, _ in enumerate(self._filenames)
        )

    def __len__(self: Self, /) -> int:
        return self._index.total_length()

    def __repr__(self: Self, /) -> str:
        return f"{type(self).__name__}({self._path})"

    def __reversed__(self: Self, /) -> Iterator[T]:
        return chain.from_iterable(
            reversed(self._cache_chunk(~i))
            for i, _ in enumerate(self._filenames)
        )

    @overload
    def __setitem__(self: Self, index: int, value: T, /) -> None: ...

    @overload
    def __setitem__(self: Self, index: slice, value: Iterable[T], /) -> None: ...

    def __setitem__(self, index, value, /):
        if isinstance(index, slice):
            raise NotImplementedError("big lists do not support slice assignments")
        index = range(len(self))[index]
        i, j = self._index.find(index)
        self._cache_chunk(i)[j] = value

    def __setstate__(self: Self, path: Path, /) -> None:
        type(self).__init__(self, path)

    def _balance(self: Self, index: int, /) -> None:
        indexer = self._index
        if len(self) == 0:
            self.clear()
            return
        elif len(self._filenames) != 1:
            pass
        elif indexer[0] > 2 * CHUNKSIZE:
            chunk = self._cache_chunk(0)
            self._filenames.append(self._get_filename())
            self._cache[self._filenames[-1]] = chunk[len(chunk) // 2:]
            indexer.update(0, -(len(chunk) + 1) // 2)
            del chunk[len(chunk) // 2 :]
            indexer.append(len(chunk))
            return
        else:
            return
        index = range(len(self._filenames))[index]
        if index == 0:
            if indexer[0] + indexer[1] < CHUNKSIZE:
                indexer.update(0, indexer[1])
                self._cache_chunk(0).extend(self._pop_chunk(1))
            elif indexer[0] + indexer[1] > 4 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(0),
                    *self._cache_chunk(1),
                ]
                self._cache_chunk(0)[:] = chunk[: len(chunk) // 3]
                indexer[0] = len(self._cache_chunk(0))
                self._cache_chunk(1)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                indexer[1] = len(self._cache_chunk(1))
                del chunk[: 2 * len(chunk) // 3]
                self._filenames.insert(2, self._get_filename())
                self._free_cache()
                self._cache[self._filenames[2]] = chunk
                indexer.insert(2, len(chunk))
            elif (
                CHUNKSIZE // 2 < indexer[0] < CHUNKSIZE * 2
                and CHUNKSIZE_EXTENDED < indexer[0] + indexer[1] < 3 * CHUNKSIZE
            ):
                pass
            elif indexer[0] > indexer[1]:
                diff = indexer[0] - indexer[1]
                self._cache_chunk(1)[:0] = self._cache_chunk(0)[-diff // 2 :]
                del self._cache_chunk(0)[-diff // 2:]
                indexer.update(0, len(self._cache_chunk(0)) - indexer[0])
                indexer.update(1, len(self._cache_chunk(1)) - indexer[1])
            else:
                diff = indexer[1] - indexer[0]
                self._cache_chunk(0).extend(self._cache_chunk(1)[: diff // 2])
                del self._cache_chunk(1)[:diff // 2]
                indexer.update(0, len(self._cache_chunk(0)) - indexer[0])
                indexer.update(1, len(self._cache_chunk(1)) - indexer[1])
        elif index + 1 == len(self._filenames):
            if indexer[-1] + indexer[-2] < CHUNKSIZE:
                indexer.update(-2, indexer[-1])
                self._cache_chunk(index - 1).extend(self._pop_chunk(index))
            elif indexer[-1] + indexer[-2] > 4 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(-2),
                    *self._cache_chunk(-1),
                ]
                self._cache_chunk(-2)[:] = chunk[: len(chunk) // 3]
                indexer[-2] = len(self._cache_chunk(-2))
                self._cache_chunk(-1)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                indexer[-1] = len(self._cache_chunk(-1))
                del chunk[: 2 * len(chunk) // 3]
                self._filenames.append(self._get_filename())
                self._free_cache()
                self._cache[self._filenames[-1]] = chunk
                indexer.append(len(chunk))
            elif (
                CHUNKSIZE // 2 < indexer[-1] < CHUNKSIZE * 2
                and CHUNKSIZE_EXTENDED < indexer[-1] + indexer[-2] < 3 * CHUNKSIZE
            ):
                pass
            elif indexer[-1] < indexer[-2]:
                diff = indexer[-2] - indexer[-1]
                self._cache_chunk(-1)[:0] = self._cache_chunk(-2)[-diff // 2 :]
                del self._cache_chunk(-2)[-diff // 2 :]
                indexer.update(-1, len(self._cache_chunk(-1)) - indexer[-1])
                indexer.update(-2, len(self._cache_chunk(-2)) - indexer[-2])
            else:
                diff = indexer[-1] - indexer[-2]
                self._cache_chunk(-2).extend(self._cache_chunk(-1)[: diff // 2])
                del self._cache_chunk(-1)[: diff // 2]
                indexer.update(-1, len(self._cache_chunk(-1)) - indexer[-1])
                indexer.update(-2, len(self._cache_chunk(-2)) - indexer[-2])
        else:
            if indexer[index - 1] + indexer[index] + indexer[index + 1] < CHUNKSIZE_EXTENDED:
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._pop_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 2]
                self._cache_chunk(index)[:] = chunk[len(chunk) // 2 :]
                indexer[index - 1] = len(chunk) // 2
                indexer[index] = (len(chunk) + 1) // 2
            elif indexer[index - 1] + indexer[index] + indexer[index + 1] > 6 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._cache_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 4]
                indexer[index - 1] = len(self._cache_chunk(index - 1))
                self._cache_chunk(index)[:] = chunk[len(chunk) // 4 : len(chunk) // 2]
                indexer[index] = len(self._cache_chunk(index))
                self._cache_chunk(index + 1)[:] = chunk[len(chunk) // 2 : 3 * len(chunk) // 4]
                indexer[index + 1] = len(self._cache_chunk(index + 1))
                del chunk[: 3 * len(chunk) // 4]
                self._filenames.insert(index + 2, self._get_filename())
                self._free_cache()
                self._cache[self._filenames[index + 2]] = chunk
                indexer.insert(index + 2, len(chunk))
            elif not all(CHUNKSIZE // 2 < 2 * L // 3 < CHUNKSIZE for L in indexer[index - 1 : index + 2]):
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._cache_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 3]
                self._cache_chunk(index)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                self._cache_chunk(index + 1)[:] = chunk[2 * len(chunk) // 3 :]
                indexer.update(index - 1, len(self._cache_chunk(index - 1)) - indexer[index - 1])
                indexer.update(index, len(self._cache_chunk(index)) - indexer[index])
                indexer.update(index + 1, len(self._cache_chunk(index + 1)) - indexer[index + 1])

    def _cache_chunk(self: Self, index: int, /) -> list[T]:
        filename = self._filenames[index]
        if filename in self._cache:
            self._cache.move_to_end(filename)
        else:
            self._free_cache()
            with open(self._path / "list" / f"{filename}.txt", mode="rb") as file:
                self._cache[filename] = pickle.load(file)
        return self._cache[filename]

    def _commit_chunk(self: Self, filename: int, segment: list[T], /) -> None:
        with open(self._path / "list" / f"{filename}.txt", mode="wb") as file:
            pickle.dump(segment, file)

    def _del_chunk(self: Self, index: int, /) -> None:
        index = range(len(self._filenames))[index]
        filename = self._filenames.pop(index)
        (self._path / "list" / f"{filename}.txt").unlink()
        del self._index[index]
        self._cache.pop(filename, None)

    def _free_cache(self: Self, /) -> None:
        while len(self._cache) >= 4:
            self._commit_chunk(*self._cache.popitem(last=False))

    def _get_filename(self: Self, /) -> int:
        path = self._path
        with open(path / "list" / "counter.txt", mode="rb") as file:
            uid = pickle.load(file)
        with open(path / "list" / "counter.txt", mode="wb") as file:
            pickle.dump(uid + 1, file)
        with open(path / "list" / f"{uid}.txt", mode="wb") as file:
            pickle.dump([], file)
        return uid

    def _pop_chunk(self: Self, index: int, /) -> list[T]:
        filename = self._filenames[index]
        if filename in self._cache:
            segment = self._cache.pop(filename)
        else:
            with open(self._path / "list" / f"{filename}.txt", mode="rb") as file:
                segment = pickle.load(file)
        self._del_chunk(index)
        return segment

    def append(self: Self, value: T, /) -> None:
        if len(self) == 0:
            self._filenames.append(self._get_filename())
            self._cache[self._filenames[-1]] = [value]
            self._index.append(1)
            return
        else:
            self._cache_chunk(-1).append(value)
            self._index.update(-1, 1)
            self._balance(-1)

    def clear(self: Self, /) -> None:
        path = self._path
        for filename in self._filenames:
            (path / "list" / f"{filename}.txt").unlink()
        with open(path / "list" / "counter.txt", mode="wb") as file:
            pickle.dump(0, file)
        with open(path / "list" / "filenames.txt", mode="wb") as file:
            pickle.dump([], file)
        with open(path / "list" / "lens.txt", mode="wb") as file:
            pickle.dump([], file)
        self._cache.clear()
        self._filenames.clear()
        self._index.clear()

    def commit(self: Self, /) -> None:
        path = self._path
        with open(path / "list" / "filenames.txt", mode="wb") as file:
            pickle.dump(self._filenames, file)
        with open(path / "list" / "lens.txt", mode="wb") as file:
            pickle.dump(self._index._data, file)
        for filename, segment in self._cache.items():
            self._commit_chunk(filename, segment)

    def extend(self: Self, iterable: Iterable[T], /) -> None:
        filenames = self._filenames
        indexer = self._index
        if not isinstance(iterable, Iterable):
            raise TypeError(f"extend expected iterable, got {iterable!r}")
        elif isinstance(iterable, list):
            if len(iterable) == 0:
                return
            elif len(self._filenames) == 1 and indexer[0] < CHUNKSIZE_EXTENDED:
                offset = CHUNKSIZE_EXTENDED - indexer[0]
                self._cache_chunk(0).extend(iterable[:offset])
                offset -= CHUNKSIZE_EXTENDED - len(self._cache_chunk(0))
                indexer.update(0, len(self._cache_chunk(0)) - indexer[0])
            else:
                offset = 0
            for i in range(offset, len(iterable), CHUNKSIZE_EXTENDED):
                chunk = iterable[i : i + CHUNKSIZE_EXTENDED]
                filenames.append(self._get_filename())
                self._commit_chunk(filenames[-1], chunk)
                indexer.append(len(chunk))
        else:
            iterator = iter(iterable)
            if len(self._filenames) == 1 and indexer[0] < CHUNKSIZE_EXTENDED:
                self._cache_chunk(0).extend(islice(iterator, CHUNKSIZE_EXTENDED - indexer[0]))
                indexer.update(0, len(self._cache_chunk(0)) - indexer[0])
            while True:
                chunk = [*islice(iterator, CHUNKSIZE_EXTENDED)]
                if not chunk:
                    break
                filenames.append(self._get_filename())
                self._commit_chunk(filenames[-1], chunk)
                indexer.append(len(chunk))

    def insert(self: Self, index: int, value: T, /) -> None:
        index = operator.index(index)
        if len(self._filenames) == 0:
            return self.append(value)
        elif index < 0:
            index += len(self)
            if index < 0:
                index = 0
        elif index >= len(self):
            return self.append(value)
        i, j = self._index.find(index)
        self._cache_chunk(i).insert(j, value)
        self._index.update(i, 1)
        self._balance(i)

    def reverse(self: Self, /) -> None:
        for i, _ in enumerate(self._filenames):
            self._cache_chunk(i).reverse()
        self._filenames.reverse()
        self._index.reverse()



from typing import Final, Generic, Optional, TypeVar, Union

from .viewable_mutable_sequence import ViewableMutableSequence
from .sequence_islice import SequenceIslice

T = TypeVar("T")

Self = TypeVar("Self", bound="MutableSequenceIslice")


class MutableSequenceIslice(SequenceIslice[T], Generic[T]):
    _sequence: Final[ViewableMutableSequence[T]]

    __slots__ = ()

    def __init__(
        self: Self,
        sequence: ViewableMutableSequence[T],
        start: Optional[int],
        stop: Optional[int],
        step: Optional[int],
        /,
    ) -> None:
        assert isinstance(sequence, ViewableMutableSequence)
        super().__init__(sequence, start, stop, step)

    def __delitem__(self: Self, index: Union[int, slice], /) -> None:
        if isinstance(index, slice):
            range_ = range(len(self._sequence))[index]
            start = None if index.start is None else range_.start
            stop = None if index.stop is None else range_.stop
            step = None if index.step is None else range_.step
            del self._sequence[start:stop:step]
        else:
            index = range(len(self._sequence))[self._start:self._stop:self._step][index]
            del self._sequence[index]



from abc import ABC, abstractmethod
from collections.abc import Iterator, Sequence
from typing import Generic, Optional, TypeVar, overload

from .sequence_islice_proxy import SequenceIsliceProxy

T_co = TypeVar("T_co", covariant=True)

Self = TypeVar("Self", bound="ViewableSequence")


class ViewableSequence(Sequence[T_co], ABC, Generic[T_co]):

    __slots__ = ()

    @overload
    def __getitem__(self: Self, index: int, /) -> T_co: ...

    @overload
    def __getitem__(self: Self, index: slice, /) -> "ViewableSequence[T_co]": ...

    @abstractmethod
    def __getitem__(self, index, /):
        raise NotImplementedError(f"__getitem__ is a required method for viewable sequences")

    def __islice__(self: Self, start: Optional[int], stop: Optional[int], step: Optional[int], /) -> Iterator[T_co]:
        len_ = len(self)
        range_ = range(len_)[start:stop:step]
        if len(range_) == 0:
            return (self[i] for i in range_)
        elif range_.step == 1 and len(range_) == len_:
            return iter(self)
        elif range_.step == 1 and range_.start == 0:
            return islice(self, range_.stop)
        elif range_.step == 1 and range_.start < range_.stop // 2:
            return islice(self, range_.start, range_.stop)
        elif range_.step == -1 and len(range_) == len_:
            return reversed(self)
        elif range_.step == -1 and range_.start + 1 == len_:
            return islice(reversed(self), len(range_))
        elif range_.step == -1 and (len_ - range_.start) < (len_ - range_.stop) // 2:
            return islice(reversed(self), len_ - range_.start - 1, len_ - range_.stop - 1)
        elif len(range_) < len_ // 8:
            return (self[i] for i in range_)
        elif range_.step > 0:
            return islice(self, range_.start, range_.stop, range_.step)
        else:
            return islice(reversed(self), len_ - range_.start - 1, len_ - range_.stop - 1, -range_.step)

    @property
    def islice(self: Self, /) -> SequenceIsliceProxy[T_co]:
        return SequenceIsliceProxy(self)



import pickle
from bisect import bisect
from collections import OrderedDict
from collections.abc import Set as AbstractSet, Iterable, Iterator
from collections.abc import MutableSet
from itertools import chain, groupby, islice
from operator import length_hint
from os import PathLike
from pathlib import Path
from types import TracebackType
from typing import Any, Final, Generic, Optional, SupportsIndex, Type
from typing import TypeVar, Union, overload

ET = TypeVar("ET", bound=BaseException)
T = TypeVar("T")

Self = TypeVar("Self", bound="BigSortedSet")

CHUNKSIZE = 1024
CHUNKSIZE_EXTENDED = 1536

assert CHUNKSIZE * 3 == CHUNKSIZE_EXTENDED * 2

def ensure_file(path: Path, default: T) -> T:
    if path.exists():
        with open(path, mode="rb") as file:
            return pickle.load(file)
    else:
        with open(path, mode="wb") as file:
            pickle.dump(default, file)
        return default


class BigSortedSet(MutableSet[T], Generic[T]):
    _cache: Final[OrderedDict[str, list[T]]]
    _filenames: Final[list[Path]]
    _len: int
    _lens: Final[list[int]]
    _mins: Final[list[T]]
    _path: Final[Path]

    __slots__ = {
        "_cache":
            "A cache containing the 16 most recently accessed segments.",
        "_filenames":
            "The file names for each segment.",
        "_len":
            "The length of the entire set.",
        "_lens":
            "The lengths of every segment.",
        "_mins":
            "The first element of each segment.",
        "_path":
            "The folder containing all of the files.",
    }

    def __init__(self: Self, path: Union[PathLike, bytes, str], /) -> None:
        path = Path(path).resolve()
        path.mkdir(exist_ok=True)
        (path / "sorted_set").mkdir(exist_ok=True)
        ensure_file(path / "sorted_set" / "counter.txt", 0)
        self._cache = OrderedDict()
        self._filenames = ensure_file(path / "sorted_set" / "filenames.txt", [])
        self._lens = ensure_file(path / "sorted_set" / "lens.txt", [])
        self._len = sum(self._lens)
        self._mins = ensure_file(path / "sorted_set" / "mins.txt", [])
        self._path = path

    def __contains__(self: Self, element: Any, /) -> bool:
        try:
            i = bisect(self._mins, element)
        except TypeError:
            return False
        if i == 0:
            return False
        j = bisect(self._cache_chunk(i - 1), element)
        return j > 0 and self._cache_chunk(i - 1)[j - 1] == element

    def __delitem__(self: Self, element: T, /) -> T:
        i = bisect(self._mins, element)
        if i == 0:
            raise KeyError(element)
        j = bisect(self._cache_chunk(i - 1), element)
        if j == 0 or element != self._cache_chunk(i - 1)[j - 1]:
            raise KeyError(element)
        del self._cache_chunk(i - 1)[j - 1]
        self._len -= 1
        self._lens[i - 1] -= 1
        if j == 1 and self._lens[i - 1] > 0:
            self._mins[i - 1] = self._cache_chunk(i - 1)[0]
        self._balance(i - 1)

    def __enter__(self: Self, /) -> Self:
        return self

    def __exit__(
        self: Self,
        exc_type: Optional[Type[ET]],
        exc_val: Optional[ET],
        exc_traceback: Optional[TracebackType],
        /,
    ) -> None:
        self.commit()

    def __getstate__(self: Self, /) -> Path:
        return self._path

    def __iter__(self: Self, /) -> Iterator[T]:
        return chain.from_iterable(
            self._cache_chunk(i)
            for i, _ in enumerate(self._filenames)
        )

    def __len__(self: Self, /) -> int:
        return self._len

    def __repr__(self: Self, /) -> str:
        return f"{type(self).__name__}({self._path})"

    def __setstate__(self: Self, path: Path, /) -> None:
        type(self).__init__(self, path)

    def _balance(self: Self, index: int, /) -> None:
        lens = self._lens
        mins = self._mins
        if self._len == 0:
            self.clear()
            return
        elif len(lens) != 1:
            pass
        elif lens[0] > 2 * CHUNKSIZE:
            chunk = self._cache_chunk(0)
            self._filenames.append(self._get_filename())
            self._cache[self._filenames[-1]] = chunk[len(chunk) // 2:]
            lens[0] = len(chunk) // 2
            del len[chunk // 2 :]
            lens.append(len(chunk))
            mins.append(chunk[0])
        else:
            return
        index = range(len(self._filenames))[index]
        if index == 0:
            if lens[0] + lens[1] < CHUNKSIZE:
                lens[0] += lens[1]
                self._len += lens[1]
                self._cache_chunk(0).extend(self._pop_chunk(1))
            elif lens[0] + lens[1] > 4 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(0),
                    *self._cache_chunk(1),
                ]
                self._cache_chunk(0)[:] = chunk[: len(chunk) // 3]
                lens[0] = len(self._cache_chunk(0))
                self._cache_chunk(1)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                lens[1] = len(self._cache_chunk(1))
                mins[1] = self._cache_chunk(1)[0]
                del chunk[: 2 * len(chunk) // 3]
                self._filenames.insert(2, self._get_filename())
                self._free_cache()
                self._cache[self._filenames[2]] = chunk
                lens.insert(2, len(chunk))
                mins.insert(2, chunk[0])
            elif (
                CHUNKSIZE // 2 < lens[0] < CHUNKSIZE * 2
                and CHUNKSIZE_EXTENDED < lens[0] + lens[1] < 3 * CHUNKSIZE
            ):
                pass
            elif lens[0] > lens[1]:
                diff = lens[0] - lens[1]
                self._cache_chunk(1)[:0] = self._cache_chunk(0)[-diff // 2 :]
                lens[1] = len(self._cache_chunk(1))
                mins[1] = self._cache_chunk(1)[0]
                del self._cache_chunk(0)[-diff // 2 :]
                lens[0] = len(self._cache_chunk(0))
            else:
                diff = lens[1] - lens[0]
                self._cache_chunk(0).extend(self._cache_chunk(1)[: diff // 2])
                lens[0] = len(self._cache_chunk(0))
                del self._cache_chunk(1)[: diff // 2]
                lens[1] = len(self._cache_chunk(1))
                mins[1] = self._cache_chunk(1)[0]
        elif index + 1 == len(self._filenames):
            if lens[-1] + lens[-2] < CHUNKSIZE:
                lens[-2] += lens[-1]
                self._cache_chunk(index - 1).extend(self._pop_chunk(index))
            elif lens[-1] + lens[-2] > 4 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(-2),
                    *self._cache_chunk(-1),
                ]
                self._cache_chunk(-2)[:] = chunk[: len(chunk) // 3]
                lens[-2] = len(self._cache_chunk(-2))
                self._cache_chunk(-1)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                lens[-1] = len(self._cache_chunk(-1))
                del chunk[: 2 * len(chunk) // 3]
                self._filenames.append(self._get_filename())
                self._free_cache()
                self._cache[self._filenames[-1]] = chunk
                lens.append(len(chunk))
            elif (
                CHUNKSIZE // 2 < lens[-1] < CHUNKSIZE * 2
                and CHUNKSIZE_EXTENDED < lens[-1] + lens[-2] < 3 * CHUNKSIZE
            ):
                pass
            elif lens[-1] < lens[-2]:
                diff = lens[-2] - lens[-1]
                self._cache_chunk(-1)[:0] = self._cache_chunk(-2)[-diff // 2 :]
                lens[-1] = len(self._cache_chunk(-1))
                del self._cache_chunk(-2)[-diff // 2 :]
                lens[-2] = len(self._cache_chunk(-2))
            else:
                diff = lens[-1] - lens[-2]
                self._cache_chunk(-2).extend(self._cache_chunk(-1)[: diff // 2])
                lens[-2] = len(self._cache_chunk(-2))
                del self._cache_chunk(-1)[: diff // 2]
                lens[-1] = len(self._cache_chunk(-1))
        else:
            if lens[index - 1] + lens[index] + lens[index + 1] < CHUNKSIZE_EXTENDED:
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._pop_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 2]
                self._cache_chunk(index)[:] = chunk[len(chunk) // 2 :]
                lens[index - 1] = len(chunk) // 2
                lens[index] = (len(chunk) + 1) // 2
            elif lens[index - 1] + lens[index] + lens[index + 1] > 6 * CHUNKSIZE:
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._cache_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 4]
                lens[index - 1] = len(self._cache_chunk(index - 1))
                self._cache_chunk(index)[:] = chunk[len(chunk) // 4 : len(chunk) // 2]
                lens[index] = len(self._cache_chunk(index))
                self._cache_chunk(index + 1)[:] = chunk[len(chunk) // 2 : 3 * len(chunk) // 4]
                lens[index + 1] = len(self._cache_chunk(index + 1))
                del chunk[: 3 * len(chunk) // 4]
                self._filenames.insert(index + 2, self._get_filename())
                self._free_cache()
                self._cache[self._filenames[index + 2]] = chunk
                lens.insert(index + 2, len(chunk))
            elif not all(CHUNKSIZE // 2 < 2 * L // 3 < CHUNKSIZE for L in lens[index - 1 : index + 2]):
                chunk = [
                    *self._cache_chunk(index - 1),
                    *self._cache_chunk(index),
                    *self._cache_chunk(index + 1),
                ]
                self._cache_chunk(index - 1)[:] = chunk[: len(chunk) // 3]
                lens[index - 1] = len(self._cache_chunk(index - 1))
                self._cache_chunk(index)[:] = chunk[len(chunk) // 3 : 2 * len(chunk) // 3]
                lens[index] = len(self._cache_chunk(index))
                self._cache_chunk(index + 1)[:] = chunk[2 * len(chunk) // 3 :]
                lens[index + 1] = len(self._cache_chunk(index + 1))

    def _cache_chunk(self: Self, index: int, /) -> list[T]:
        filename = self._filenames[index]
        if filename in self._cache:
            self._cache.move_to_end(filename)
        else:
            self._free_cache()
            with open(self._path / "sorted_set" / f"{filename}.txt", mode="rb") as file:
                self._cache[filename] = pickle.load(file)
        return self._cache[filename]

    def _commit_chunk(self: Self, filename: int, segment: list[T], /) -> None:
        with open(self._path / "sorted_set" / f"{filename}.txt", mode="wb") as file:
            pickle.dump(segment, file)

    def _del_chunk(self: Self, index: int, /) -> None:
        index = range(len(self._filenames))[index]
        filename = self._filenames.pop(index)
        (self._path / "sorted_set" / f"{filename}.txt").unlink()
        self._len -= self._lens.pop(index)
        del self._mins[index]

    def _free_cache(self: Self, /) -> None:
        while len(self._cache) >= 16:
            self._commit_chunk(*self._cache.popitem(last=False))

    def _get_filename(self: Self, /) -> int:
        path = self._path
        with open(path / "sorted_set" / "counter.txt", mode="rb") as file:
            uid = pickle.load(file)
        with open(path / "sorted_set" / "counter.txt", mode="wb") as file:
            pickle.dump(uid + 1, file)
        with open(path / "sorted_set" / f"{uid}.txt", mode="wb") as file:
            pickle.dump([], file)
        return uid

    def _pop_chunk(self: Self, index: int, /) -> list[T]:
        filename = self._filenames[index]
        if filename in self._cache:
            segment = self._cache.pop(filename)
        else:
            with open(self._path / "sorted_set" / f"{filename}.txt", mode="rb") as file:
                segment = pickle.load(file)
        self._del_chunk(index)
        return segment

    def add(self: Self, element: T, /) -> None:
        i = bisect(self._mins, element)
        if i > 0:
            j = bisect(self._cache_chunk(i - 1), element)
            if j == 0 or self._cache_chunk(i - 1)[j - 1] != element:
                self._cache_chunk(i - 1).insert(j, element)
                self._len += 1
                self._lens[i - 1] += 1
                self._balance(i - 1)
        elif self._len > 0:
            self._cache_chunk(0).insert(0, element)
            self._len += 1
            self._lens[0] += 1
            self._balance(0)
        else:
            self._filenames.append(self._get_filename())
            self._cache[self._filenames[-1]] = [element]
            self._len = 1
            self._lens.append(1)

    def clear(self: Self, /) -> None:
        path = self._path
        for filename in self._filenames:
            (path / "sorted_set" / f"{filename}.txt").unlink()
        with open(path / "sorted_set" / "counter.txt", mode="wb") as file:
            pickle.dump(0, file)
        with open(path / "sorted_set" / "filenames.txt", mode="wb") as file:
            pickle.dump([], file)
        with open(path / "sorted_set" / "lens.txt", mode="wb") as file:
            pickle.dump([], file)
        with open(path / "sorted_set" / "mins.txt", mode="wb") as file:
            pickle.dump([], file)
        self._cache.clear()
        self._filenames.clear()
        self._len = 0
        self._lens.clear()
        self._mins.clear()

    def commit(self: Self, /) -> None:
        path = self._path
        with open(path / "sorted_set" / "filenames.txt", mode="wb") as file:
            pickle.dump(self._filenames, file)
        with open(path / "sorted_set" / "lens.txt", mode="wb") as file:
            pickle.dump(self._lens, file)
        with open(path / "sorted_set" / "mins.txt", mode="wb") as file:
            pickle.dump(self._mins, file)
        for filename, segment in self._cache.items():
            self._commit_chunk(filename, segment)

    def difference_update(self: Self, /, *iterables: Iterable[Any]) -> None:
        for iterable in iterables:
            for element in iterable:
                self.discard(element)

    def discard(self: Self, element: Any, /) -> None:
        i = bisect(self._mins, element)
        if i > 0:
            j = bisect(self._cache_chunk(i - 1), element)
            if j > 0 and self._cache_chunk(i - 1)[j - 1] == element:
                del self._cache_chunk(i - 1)[j]
                self._len -= 1
                self._lens[i - 1] -= 1
                self._balance(i - 1)

    def intersection_update(self: Self, /, *iterables: Iterable[Any]) -> None:
        path = self._path
        temp = path / "__temp__"
        for iterable in iterables:
            with BigSortedSet(temp) as db:
                db.update(element for element in iterable if element in self)
            for filename in ("counter.txt", "filenames.txt", "lens.txt", "mins.txt"):
                (temp / "sorted_set" / filename).rename(
                    path / "sorted_set" / filename
                )
            for filename in db._filenames:
                (temp / "sorted_set" / f"{filename}.txt").rename(
                    path / "sorted_set" / f"{filename}.txt"
                )
            self._filenames[:] = db._filenames[:]
            self._len = db._len
            self._lens[:] = db._lens[:]
            self._mins[:] = db._mins[:]

    def isdisjoint(self: Self, iterable: Iterable[Any], /) -> bool:
        if length_hint(iterable) > len(self) and isinstance(iterable, AbstractSet):
            return any(element in iterable for element in self)
        else:
            return any(element in self for element in iterable)

    def issubset(self: Self, iterable: Iterable[Any], /) -> bool:
        if isinstance(iterable, AbstractSet):
            return iterable >= self
        path = self._path
        temp = path / "__temp__"
        with BigSortedSet(temp) as db:
            db.update(element for element in iterable if element in self)
            result = len(db) == len(self)
            db.clear()
        return result

    def issuperset(self: Self, iterable: Iterable[Any], /) -> bool:
        return all(element in self for element in iterable)

    def pop(self: Self, /) -> T:
        if self._len == 0:
            raise KeyError
        else:
            self._len -= 1
            self._lens[-1] -= 1
            element = self._cache_chunk(-1).pop()
            self._balance(-1)
            return element

    def remove(self: Self, element: Any, /) -> None:
        len_ = self._len
        self.discard(element)
        if self._len == len_:
            raise KeyError(element)

    def symmetric_difference_update(self: Self, /, *iterables: Iterable[Any]) -> None:
        path = self._path
        temp = path / "__temp__"
        for iterable in iterables:
            with BigSortedSet(temp) as db:
                db.update(iterable)
                for _ in reversed(range(len(db._filenames))):
                    for element in db._pop_chunk(-1):
                        len_ = self._len
                        self.add(element)
                        if self._len == len_:
                            self.remove(element)
                db.clear()

    def update(self: Self, /, *iterables: Iterable[Any]) -> None:
        iterator = chain.from_iterable(iterables)
        if self._len == 0:
            filenames = self._filenames
            front = (key for key, _ in groupby(sorted(islice(iterator, CHUNKSIZE_EXTENDED ** 2))))
            while True:
                chunk = [*islice(front, CHUNKSIZE_EXTENDED)]
                if not chunk:
                    return
                filenames.append(self._get_filename())
                self._commit_chunk(filenames[-1], chunk)
                self._len += len(chunk)
                self._lens.append(len(chunk))
                self._mins.append(chunk[0])
        for element in iterator:
            self.add(element)



from abc import ABC, abstractmethod
from collections.abc import MutableSequence
from typing import Generic, TypeVar, overload

from .mutable_sequence_islice_proxy import MutableSequenceIsliceProxy
from .viewable_sequence import ViewableSequence

T = TypeVar("T")

Self = TypeVar("Self", bound="ViewableMutableSequence")


class ViewableMutableSequence(ViewableSequence[T], MutableSequence[T], ABC, Generic[T]):

    __slots__ = ()

    @overload
    def __getitem__(self: Self, index: int, /) -> T: ...

    @overload
    def __getitem__(self: Self, index: slice, /) -> "ViewableMutableSequence[T]": ...

    @abstractmethod
    def __getitem__(self, index, /):
        raise NotImplementedError(f"__getitem__ is a required method for viewable mutable sequences")

    @property
    def islice(self: Self, /) -> MutableSequenceIsliceProxy[T]:
        return MutableSequenceIsliceProxy(self)



from collections.abc import Iterator
from typing import Final, Generic, TypeVar

import advanced.databases._src as src

T_co = TypeVar("T_co", covariant=True)

Self = TypeVar("Self", bound="SequenceIsliceProxy")


class SequenceIsliceProxy(Generic[T_co]):
    _sequence: Final["src.viewable_sequence.ViewableSequence[T_co]"]

    __slots__ = {
        "_sequence":
            "The viewable sequence.",
    }

    def __init__(self: Self, sequence: "src.viewable_sequence.ViewableSequence[T_co]", /) -> None:
        assert isinstance(sequence, src.viewable_sequence.ViewableSequence)
        self._sequence = sequence

    def __getitem__(self: Self, index: slice, /) -> "src.sequence_islice.SequenceIslice[T_co]":
        if not isinstance(index, slice):
            raise TypeError(f"expected sequence.islice[start:stop:step], got sequence.islice[{index!r}]")
        # Cast integer-like indices to integers.
        range_ = range(len(self._sequence))[index]
        # Collect None or integers.
        start = None if index.start is None else range_.start
        stop = None if index.stop is None else range_.stop
        step = None if index.step is None else range_.step
        return src.sequence_islice.SequenceIslice(self._sequence, start, stop, step)



from typing import Generic, TypeVar, Union, overload

import advanced.databases._src as src
from .sequence_islice_proxy import SequenceIsliceProxy

T = TypeVar("T")

Self = TypeVar("Self", bound="MutableSequenceIsliceProxy")


class MutableSequenceIsliceProxy(SequenceIsliceProxy[T], Generic[T]):
    _sequence: "src.viewable_mutable_sequence.ViewableMutableSequence[T]"

    __slots__ = ()

    def __init__(self: Self, sequence: "src.viewable_mutable_sequence.ViewableMutableSequence[T]", /) -> None:
        assert isinstance(sequence, src.viewable_mutable_sequence.ViewableMutableSequence)
        self._sequence = sequence

    def __getitem__(self: Self, index: slice, /) -> "src.mutable_sequence_islice.MutableSequenceIslice[T]":
        if not isinstance(index, slice):
            raise TypeError(f"expected sequence.islice[start:stop:step], got sequence.islice[{index!r}]")
        # Cast integer-like indices to integers.
        range_ = range(len(self._sequence))[index]
        # Collect None or integers.
        start = None if index.start is None else range_.start
        stop = None if index.stop is None else range_.stop
        step = None if index.step is None else range_.step
        return src.mutable_sequence_islice.MutableSequenceIslice(self._sequence, start, stop, step)



from __future__ import annotations
from typing import Final, Generic, Optional, TypeVar, overload

from .viewable_sequence import ViewableSequence

T_co = TypeVar("T_co", covariant=True)

Self = TypeVar("Self", bound="SequenceIslice")


class SequenceIslice(ViewableSequence[T_co], Generic[T_co]):
    _sequence: Final[ViewableSequence[T_co]]
    _start: Optional[int]
    _stop: Optional[int]
    _step: Optional[int]

    __slots__ = {
        "_sequence":
            "The viewable sequence.",
        "_start":
            "The starting index.",
        "_stop":
            "The stopping index.",
        "_step":
            "The step size.",
    }

    def __init__(
        self: Self,
        sequence: ViewableSequence[T_co],
        start: Optional[int],
        stop: Optional[int],
        step: Optional[int],
        /,
    ) -> None:
        assert isinstance(sequence, ViewableSequence)
        self._sequence = sequence
        self._start = start
        self._stop = stop
        self._step = step

    @overload
    def __getitem__(self: Self, index: int, /) -> T_co: ...

    @overload
    def __getitem__(self: Self, index: slice, /) -> SequenceIslice[T_co]: ...

    def __getitem__(self, index, /):
        if isinstance(index, slice):
            range_ = range(len(self._sequence))[index]
            start = None if index.start is None else range_.start
            stop = None if index.stop is None else range_.stop
            step = None if index.step is None else range_.step
            return type(self)(self, start, stop, step)
        else:
            index = range(len(self._sequence))[self._start:self._stop:self._step][index]
            return self._sequence[index]

    def __islice__(self: Self, start: Optional[int], stop: Optional[int], step: Optional[int], /) -> Iterator[T_co]:
        range_ = range(len(self._sequence))[self._start:self._stop:self._step][start:stop:step]
        return self._sequence.__islice__(range_.start, range_.stop, range_.step)

    def __iter__(self: Self, /) -> Iterator[T_co]:
        return self._sequence.__islice__(self._start, self._stop, self._step)

    def __len__(self: Self, /) -> int:
        return len(range(len(self._sequence))[self._start:self._stop:self._step])

    def __repr__(self: Self, /) -> str:
        start = "" if self._start is None else repr(self._start)
        stop = "" if self._stop is None else repr(self._stop)
        step = "" if self._step is None else f":{self._step!r}"
        return f"{self._sequence!r}[{start}:{stop}{step}]"

    def __reversed__(self: Self, /) -> Iterator[T_co]:
        return self.__islice__(None, None, -1)



import operator
from collections.abc import Iterable, Iterator, MutableSequence, Sequence
from itertools import accumulate, groupby
from typing import Optional, Union, overload


class FenwickIndex(MutableSequence[int]):
    """
    A fenwick tree used to track nested indexing. Useful
    for segmented arrays or implicitly segmented arrays.

    Examples
    --------
    Tracking elements inside of nested data:
        >>> data = [[0, 1, 2, 3], [4, 5, 6]]
        >>> indexes = FenwickIndex(map(len, data))
        >>> 
        >>> indexes.find(3)
        (0, 3)
        >>> data[0][3]
        3
        >>> 
        >>> indexes.find(5)
        (1, 1)
        >>> data[1][1]
        5
        >>> 
        >>> data[0].append(-1)
        >>> indexes.update(0, +1)
        >>> 
        >>> indexes.find(5)
        (1, 0)
        >>> data[1][0]
        4
        >>> 
        >>> del data[0][-1]
        >>> indexes.update(0, -1)
        >>> 
        >>> indexes.find(5)
        (1, 1)
        >>> data[1][1]
        5

    Implicitly nested data inside of flat data:
        >>> data = [0, 1, 2, 3, 4, 5, 6]
        >>> # Implicitly split into 4 and 3 items.
        >>> indexes = FenwickIndex([4, 3])
        >>> 
        >>> indexes.upto(0), indexes[0]
        (0, 4)
        >>> data[0 : 0 + 4]
        [0, 1, 2, 3]
        >>> 
        >>> indexes.upto(1), indexes[1]
        (4, 3)
        >>> data[4 : 4 + 3]
        [4, 5, 6]
    """
    _data: list[int]
    _fenwick: Optional[list[int]]
    _len: Optional[int]

    __slots__ = ("_data", "_fenwick", "_len")

    def __init__(self, data: Iterable[int] = ()) -> None:
        self._data = [*map(operator.index, data)]
        self._fenwick = None
        self._len = None

    def __delitem__(self, index: Union[int, slice]) -> None:
        """
        Delete a segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> del data[0]
            >>> del indexes[0]
            >>> 
            >>> indexes
            FenwickIndex([3])
        """
        data = self._data
        if isinstance(index, slice):
            deleted = len(range(len(data))[index])
            if deleted == len(data):
                self.clear()
            elif deleted > 0:
                del data[index]
                self._fenwick = None
            if self._len is not None:
                self._len -= deleted
        else:
            index = range(len(data))[index]
            fenwick = self._fenwick
            if fenwick is not None:
                if index + 1 == len(data):
                    del fenwick[-1]
                else:
                    self._fenwick = None
            if self._len is None:
                del self._data[index]
            else:
                self._len -= self._data.pop(index)

    @overload
    def __getitem__(self, index: int) -> int: ...

    @overload
    def __getitem__(self, index: slice) -> list[int]: ...

    def __getitem__(self, index):
        """
        Get the length of a segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> indexes[0]
            4
        """
        return self._data[index]

    def __iter__(self) -> Iterator[int]:
        """
        Iterate over the length of each segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> print(*indexes)
            4 3
        """
        return iter(self._data)

    def __len__(self) -> int:
        """
        Returns the amount of segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> print(len(indexes))
            2
        """
        return len(self._data)

    def __repr__(self) -> str:
        """
        Returns a string representation of the fenwick array.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> indexes
            FenwickIndex([4, 3])
        """
        return f"{type(self).__name__}({self._data})"

    @overload
    def __setitem__(self, index: int, value: int) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[int]) -> None: ...

    def __setitem__(self, index, value):
        """
        Set the length of a segment.

        Equivalent to self.update(index, value - self[index]),
        but performs less efficiently.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> data[0].append(-1)
            >>> indexes[0] = len(data[0])
            >>> indexes[0]
            5
        """
        if isinstance(index, slice):
            self._fenwick = None
            self._data[index] = value
            self._len = None
        else:
            self.update(index, value - self[index])

    def append(self, value: int) -> None:
        """
        Append the length of a new segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> data.append([7, 8, 9])
            >>> indexes.append(3)
            >>> print(*indexes)
            4 3 3
        """
        self._data.append(value)
        if self._len is not None:
            self._len += value
        fenwick = self._fenwick
        if fenwick is not None:
            i = len(fenwick)
            j = i & -i
            fenwick.append(value)
            while j > 1:
                j //= 2
                fenwick[i] += fenwick[i - j]

    def clear(self) -> None:
        """
        Clear the segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> data.clear()
            >>> indexes.clear()
            >>> indexes
            FenwickIndex([])
        """
        self._data.clear()
        self._fenwick = None
        self._len = None

    def find(self, index: int) -> tuple[int, int]:
        """
        Find the segment and position in the segment for an index.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> indexes.find(3)
            (0, 3)
            >>> data[0][3]
            3
            >>> 
            >>> indexes.find(5)
            (1, 1)
            >>> data[1][1]
            5
        """
        if self._len is None:
            self.total_length()
        index = range(self._len)[index]
        fenwick = self._fenwick
        if fenwick is None:
            fenwick = [0, *self._data]
            fenwick_len = len(fenwick)
            for i in range(1, fenwick_len):
                j = i + (i & -i)
                if j < fenwick_len:
                    fenwick[j] += fenwick[i]
            self._fenwick = fenwick
        else:
            fenwick_len = len(self._fenwick)
        i = 0
        j = 1 << fenwick_len.bit_length()
        while j > 0:
            i += j
            if i < fenwick_len and fenwick[i] <= index:
                index -= fenwick[i]
            else:
                i -= j
            j //= 2
        return (i, index)

    def insert(self, index: int, value: int) -> None:
        """
        Insert the length of a segment before an index.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> data.insert(0, [-1] * 5)
            >>> indexes.insert(0, 5)
            >>> 
            >>> indexes
            FenwickIndex([5, 4, 3])
        """
        data = self._data
        index = operator.index(index)
        if index < 0:
            index += len(data)
        if index >= len(data):
            self.append(value)
        else:
            data.insert(index, value)
            self._fenwick = None
        if self._len is not None:
            self._len += value

    def reverse(self) -> None:
        self._data.reverse()
        self._fenwick = None

    def total_length(self) -> int:
        """
        Returns the combined length of all segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> indexes.total_length()
            7
        """
        if self._len is None:
            if len(self._data) == 0:
                self._len = 0
            else:
                self._len = self.upto(-1) + self._data[-1]
        return self._len

    def update(self, index: int, value: int) -> None:
        """
        Update the length of a segment.

        Equivalent to self[index] += value,
        but performs more efficiently.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = FenwickIndex(map(len, data))
            >>> 
            >>> data[0].append(-1)
            >>> indexes.update(0, 1)
            >>> indexes[0]
            5
        """
        index = range(len(self._data))[index]
        self._data[index] += value
        if self._len is not None:
            self._len += value
        fenwick = self._fenwick
        if fenwick is not None:
            fenwick_len = len(fenwick)
            index = range(1, fenwick_len)[index]
            if index & (index - 1) == 0:
                while index < fenwick_len:
                    fenwick[index] += value
                    index *= 2
            else:
                while index < fenwick_len:
                    fenwick[index] += value
                    index += index & -index

    def upto(self, index: int) -> int:
        """
        Find the length of all segments upto (not including) a segment.

        Example
        -------
            >>> data = [0, 1, 2, 3, 4, 5, 6]
            >>> # Implicitly split into 4 and 3 items.
            >>> indexes = FenwickIndex([4, 3])
            >>> 
            >>> indexes.upto(0), indexes[0]
            (0, 4)
            >>> data[0 : 0 + 4]
            [0, 1, 2, 3]
            >>> 
            >>> indexes.upto(1), indexes[1]
            (4, 3)
            >>> data[4 : 4 + 3]
            [4, 5, 6]
        """
        index = range(len(self._data))[index]
        fenwick = self._fenwick
        if fenwick is None:
            fenwick = [0, *self._data]
            fenwick_len = len(fenwick)
            for i in range(1, fenwick_len):
                j = i + (i & -i)
                if j < fenwick_len:
                    fenwick[j] += fenwick[i]
            self._fenwick = fenwick
        else:
            fenwick_len = len(self._fenwick)
        result = 0
        while index > 0:
            result += fenwick[index]
            index -= index & -index
        return result


class NestedIndex(MutableSequence[int]):
    """
    Tracks nested indexes where groups of consecutive segments are
    expected to have the same sizes, similar to numpy arrays but with
    more flexibility.

    Examples
    --------
    Tracking elements inside of nested data:
        >>> data = [[0, 1, 2, 3], [4, 5, 6]]
        >>> indexes = NestedIndex(map(len, data))
        >>> 
        >>> indexes.find(3)
        (0, 3)
        >>> data[0][3]
        3
        >>> 
        >>> indexes.find(5)
        (1, 1)
        >>> data[1][1]
        5
        >>> 
        >>> data[0].append(-1)
        >>> indexes.update(0, +1)
        >>> 
        >>> indexes.find(5)
        (1, 0)
        >>> data[1][0]
        4
        >>> 
        >>> del data[0][-1]
        >>> indexes.update(0, -1)
        >>> 
        >>> indexes.find(5)
        (1, 1)
        >>> data[1][1]
        5

    Implicitly nested data inside of flat data:
        >>> data = [0, 1, 2, 3, 4, 5, 6]
        >>> # Implicitly split into 4 and 3 items.
        >>> indexes = NestedIndex([4, 3])
        >>> 
        >>> indexes.upto(0), indexes[0]
        (0, 4)
        >>> data[0 : 0 + 4]
        [0, 1, 2, 3]
        >>> 
        >>> indexes.upto(1), indexes[1]
        (4, 3)
        >>> data[4 : 4 + 3]
        [4, 5, 6]
    """
    _fenwick: FenwickIndex
    _segments: FenwickIndex

    __slots__ = ("_fenwick", "_segments")

    def __init__(self, data: Iterable[int] = ()) -> None:
        fenwick = []
        segments = []
        for key, group in groupby(map(operator.index, data)):
            fenwick.append(key)
            segments.append(sum(1 for _ in group))
        self._fenwick = FenwickIndex(x * y for x, y in zip(fenwick, segments))
        self._segments = FenwickIndex(segments)

    def __delitem__(self, index: Union[int, slice]) -> None:
        """
        Delete a segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> del data[0]
            >>> del indexes[0]
            >>> 
            >>> indexes
            NestedIndex([3])
        """
        fenwick = self._fenwick
        segments = self._segments
        if isinstance(index, slice):
            range_ = range(len(segments))[index]
            if len(range_) == len(data):
                self.clear()
            elif len(range_) > 0:
                if range_.step > 0:
                    range_ = range_[::-1]
                for i in range_:
                    del self[i]
        else:
            i = segments.find(index)[0]
            if segments[i] == 1:
                del fenwick[i]
                del segments[i]
            else:
                fenwick.update(i, -fenwick[i] // segments[i])
                segments.update(i, -1)

    @overload
    def __getitem__(self, index: int) -> int: ...

    @overload
    def __getitem__(self, index: slice) -> list[int]: ...

    def __getitem__(self, index):
        """
        Get the length of a segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> indexes[0]
            4
        """
        if isinstance(index, slice):
            range_ = range(len(self))[index]
            return [self[index] for i in range_]
        else:
            i = self._segments.find(index)[0]
            return self._fenwick[i] // self._segments[i]

    def __iter__(self) -> Iterator[int]:
        """
        Iterate over the length of each segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> print(*indexes)
            4 3
        """
        return (
            x // y
            for x, y in zip(self._fenwick, self._segments)
            for _ in range(y)
        )

    def __len__(self) -> int:
        """
        Returns the amount of segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> print(len(indexes))
            2
        """
        return self._segments.total_length()

    def __repr__(self) -> str:
        """
        Returns a string representation of the fenwick array.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> indexes
            NestedIndex([4, 3])
        """
        return f"{type(self).__name__}({[*self]})"

    @overload
    def __setitem__(self, index: int, value: int) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[int]) -> None: ...

    def __setitem__(self, index, value):
        """
        Set the length of a segment.

        Equivalent to self.update(index, value - self[index]),
        but performs less efficiently.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> data[0].append(-1)
            >>> indexes[0] = len(data[0])
            >>> indexes[0]
            5
        """
        if isinstance(index, slice):
            range_ = range(len(self))[index]
            if not isinstance(value, Sequence):
                value = [*value]
            if range_.step == 1:
                iterator = iter(value)
                for i, x in zip(range_, iterator):
                    self[i] = x
                if len(range_) < len(value):
                    if range_.stop == len(self):
                        self.extend(iterator)
                    else:
                        for i, x in enumerate(iterator, range_.stop):
                            self.insert(i, x)
                elif len(range_) > len(value):
                    del self[range_.start + len(value) : range_.stop]
            elif len(range_) != len(value):
                raise ValueError(f"attempt to assign sequence of size {len(value)} to extended slice of size {len(range_)}")
            else:
                for i, x in zip(range_, value):
                    self[i] = x
        else:
            self.update(index, value - self[index])

    def append(self, value: int) -> None:
        """
        Append the length of a new segment.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> data.append([7, 8, 9])
            >>> indexes.append(3)
            >>> print(*indexes)
            4 3 3
        """
        fenwick = self._fenwick
        segments = self._segments
        if len(self) == 0:
            fenwick.append(value)
            segments.append(1)
        elif fenwick[-1] // segments[-1] == value:
            fenwick.update(-1, value)
            segments.update(-1, +1)
        else:
            fenwick.append(value)
            segments.append(1)

    def clear(self) -> None:
        """
        Clear the segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> data.clear()
            >>> indexes.clear()
            >>> indexes
            NestedIndex([])
        """
        self._fenwick.clear()
        self._segments.clear()

    def find(self, index: int) -> tuple[int, int]:
        """
        Find the segment and position in the segment for an index.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> indexes.find(3)
            (0, 3)
            >>> data[0][3]
            3
            >>> 
            >>> indexes.find(5)
            (1, 1)
            >>> data[1][1]
            5
        """
        i, j = self._fenwick.find(index)
        j, k = divmod(j, self._fenwick[i] // self._segments[i])
        return self._segments.upto(i) + j, k

    def insert(self, index: int, value: int) -> None:
        """
        Insert the length of a segment before an index.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> data.insert(0, [-1] * 5)
            >>> indexes.insert(0, 5)
            >>> 
            >>> indexes
            NestedIndex([5, 4, 3])
        """
        fenwick = self._fenwick
        segments = self._segments
        i, j = segments.find(index)
        fi = fenwick[i]
        si = segments[i]
        fs = fi // si
        if fs == value:
            fenwick.update(i, value)
            segments.update(i, value)
        elif j == 0:
            if i > 0 and fenwick[i - 1] // segments[i - 1] == value:
                fenwick.update(i - 1, value)
                segments.update(i - 1, 1)
            else:
                fenwick.insert(i, value)
                segments.insert(i, 1)
        elif j + 1 == si:
            if i + 1 < len(fenwick) and fenwick[i + 1] // segments[i + 1] == value:
                fenwick.update(i + 1, value)
                segments.update(i + 1, 1)
            elif i + 1 == len(fenwick):
                fenwick.append(value)
                segments.append(1)
            else:
                fenwick.insert(i + 1, value)
                segments.insert(i + 1, 1)
        else:
            fenwick.insert(i + 1, value)
            fenwick.insert(i + 2, fi - j * fs)
            fenwick.update(i, j * fs - fi)
            segments.insert(i + 1, 1)
            segments.insert(i + 2, si - j)
            segments.update(i, j - si)

    def reverse(self) -> None:
        self._fenwick.reverse()
        self._segments.reverse()

    def total_length(self) -> int:
        """
        Returns the combined length of all segments.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> indexes.total_length()
            7
        """
        return self._fenwick.total_length()

    def update(self, index: int, value: int) -> None:
        """
        Update the length of a segment.

        Equivalent to self[index] += value,
        but performs more efficiently.

        Example
        -------
            >>> data = [[0, 1, 2, 3], [4, 5, 6]]
            >>> indexes = NestedIndex(map(len, data))
            >>> 
            >>> data[0].append(-1)
            >>> indexes.update(0, 1)
            >>> indexes[0]
            5
        """
        index = range(len(self))[index]
        if value == 0:
            return
        fenwick = self._fenwick
        segments = self._segments
        i, j = segments.find(index)
        fi = fenwick[i]
        si = segments[i]
        fs = fi // si
        if j == 0:
            if i > 0 and fenwick[i - 1] // segments[i - 1] == fs + value:
                fenwick.update(i - 1, fs + value)
                fenwick.update(i, -fs)
                segments.update(i - 1, 1)
                segments.update(i, -1)
            else:
                fenwick.insert(i, fs + value)
                segments.insert(i, 1)
        elif j + 1 == si:
            if i + 1 < len(segments) and fenwick[i + 1] // segments[i + 1] == fs + value:
                fenwick.update(i, -fs)
                fenwick.update(i + 1, fs + value)
                segments.update(i, -1)
                segments.update(i + 1, 1)
            else:
                fenwick.insert(i + 1, fs + value)
                segments.insert(i + 1, 1)
        else:
            fenwick.insert(i + 1, fs + value)
            fenwick.insert(i + 2, fi - j * fs)
            fenwick.update(i, j * fs - fi)
            segments.insert(i + 1, 1)
            segments.insert(i + 2, si - j)
            segments.update(i, j - si)

    def upto(self, index: int) -> int:
        """
        Find the length of all segments upto (not including) a segment.

        Example
        -------
            >>> data = [0, 1, 2, 3, 4, 5, 6]
            >>> # Implicitly split into 4 and 3 items.
            >>> indexes = NestedIndex([4, 3])
            >>> 
            >>> indexes.upto(0), indexes[0]
            (0, 4)
            >>> data[0 : 0 + 4]
            [0, 1, 2, 3]
            >>> 
            >>> indexes.upto(1), indexes[1]
            (4, 3)
            >>> data[4 : 4 + 3]
            [4, 5, 6]
        """
        fenwick = self._fenwick
        segments = self._segments
        i, j = segments.find(index)
        return fenwick.upto(i) + fenwick[i] // segments[i] * j
